//
// PS2PC: An Intel x86 emulator for MIPS R5900 (Sony PS2).
// Coded by Mike Tedder / breakpoint - bpoint@parodius.com
// http://ps2pc.parodius.com
//
// This project is released as 'open source'.  You may make any modifications
// or adjustments to the code and submit a patch to be applied to the project.
// See the web page for more details.
//
//-----------------------------------------------------------------------

#define PC_VERSION			1			// 1 to enable win32 single-pass compiler

#include "types.h"
#include "x86.h"
#include "x86arith.h"
#include "x86bit.h"
#include "x86mov.h"
#include "x86stack.h"
#include "x86pmode.h"
#include "x86io.h"
#include "x86jmp.h"
#include "x86misc.h"
#include "x86fpu.h"

/*
EAX - T0
EBX - T1
ECX - T2
EDX - T3
ESI - T4
EDI - T5
EBP - T6
ESP - T7
EIP - T8
EFL - T9
CS  - S2
DS  - S3
ES  - S4
SS  - S5
FS  - S6
GS  - S7

S0 and S1 are always available as temp registers!

Supported opcodes:
  DB 66h prefix
  MOV AL..DL, imm8
  MOV AH..DH, imm8
  MOV AX..BP, imm16
  MOV EAX..EBP, imm32
  NOP
  INC AX..BP              (no flags updating!)
  INC EAX..EBP            (no flags updating!)
  DEC AX..BP              (no flags updating!)
  DEC EAX..EBP            (no flags updating!)
*/

int x86Undefined(uchar *in, ulong *out, bool *isBranch, ulong *skip);

int x86SecondByte(uchar *in, ulong *out, bool *isBranch, ulong *skip);
int x86OpSize(uchar *in, ulong *out, bool *isBranch, ulong *skip);
int x86AddrSize(uchar *in, ulong *out, bool *isBranch, ulong *skip);

int x86Grp1_EbIb(uchar *in, ulong *out, bool *isBranch, ulong *skip);
int x86Grp1_EvIv(uchar *in, ulong *out, bool *isBranch, ulong *skip);
int x86Grp1_EvIb(uchar *in, ulong *out, bool *isBranch, ulong *skip);

int x86Grp2_EbIb(uchar *in, ulong *out, bool *isBranch, ulong *skip);
int x86Grp2_EvIb(uchar *in, ulong *out, bool *isBranch, ulong *skip);
int x86Grp2_Eb1(uchar *in, ulong *out, bool *isBranch, ulong *skip);
int x86Grp2_Ev1(uchar *in, ulong *out, bool *isBranch, ulong *skip);
int x86Grp2_EbCL(uchar *in, ulong *out, bool *isBranch, ulong *skip);
int x86Grp2_EvCL(uchar *in, ulong *out, bool *isBranch, ulong *skip);

int x86Grp3_Eb(uchar *in, ulong *out, bool *isBranch, ulong *skip);
int x86Grp3_Ev(uchar *in, ulong *out, bool *isBranch, ulong *skip);

int x86Grp4(uchar *in, ulong *out, bool *isBranch, ulong *skip);
int x86Grp5(uchar *in, ulong *out, bool *isBranch, ulong *skip);


int (*x86OpcodeTable[256])(uchar *in, ulong *out, bool *isBranch, ulong *skip) = {
	x86ADD_EbGb,   x86ADD_EvGv,    x86ADD_GbEb,    x86ADD_GvEv,     x86ADD_ALIb,    x86ADD_eAXIv,    x86PUSH_ES,     x86POP_ES,			// 0xh
	x86OR_EbGb,    x86OR_EvGv,     x86OR_GbEb,     x86OR_GvEv,      x86OR_ALIb,     x86OR_eAXIv,     x86PUSH_CS,     x86SecondByte,
	x86ADC_EbGb,   x86ADC_EvGv,    x86ADC_GbEb,    x86ADC_GvEv,     x86ADC_ALIb,    x86ADC_eAXIv,    x86PUSH_SS,     x86POP_SS,			// 1xh
	x86SBB_EbGb,   x86SBB_EvGv,    x86SBB_GbEb,    x86SBB_GvEv,     x86SBB_ALIb,    x86SBB_eAXIv,    x86PUSH_DS,     x86POP_DS,
	x86AND_EbGb,   x86AND_EvGv,    x86AND_GbEb,    x86AND_GvEv,     x86AND_ALIb,    x86AND_eAXIv,    x86Seg_ES,      x86DAA,			// 2xh
	x86SUB_EbGb,   x86SUB_EvGv,    x86SUB_GbEb,    x86SUB_GvEv,     x86SUB_ALIb,    x86SUB_eAXIv,    x86Seg_CS,      x86DAS,
	x86XOR_EbGb,   x86XOR_EvGv,    x86XOR_GbEb,    x86XOR_GvEv,     x86XOR_ALIb,    x86XOR_eAXIv,    x86Seg_SS,      x86AAA,			// 3xh
	x86CMP_EbGb,   x86CMP_EvGv,    x86CMP_GbEb,    x86CMP_GvEv,     x86CMP_ALIb,    x86CMP_eAXIv,    x86Seg_DS,      x86AAS,
	x86INC_eAX,    x86INC_eCX,     x86INC_eDX,     x86INC_eBX,      x86INC_eSP,     x86INC_eBP,      x86INC_eSI,     x86INC_eDI,		// 4xh
	x86DEC_eAX,    x86DEC_eCX,     x86DEC_eDX,     x86DEC_eBX,      x86DEC_eSP,     x86DEC_eBP,      x86DEC_eSI,     x86DEC_eDI,
	x86PUSH_eAX,   x86PUSH_eCX,    x86PUSH_eDX,    x86PUSH_eBX,     x86PUSH_eSP,    x86PUSH_eBP,     x86PUSH_eSI,    x86PUSH_eDI,		// 5xh
	x86POP_eAX,    x86POP_eCX,     x86POP_eDX,     x86POP_eBX,      x86POP_eSP,     x86POP_eBP,      x86POP_eSI,     x86POP_eDI,
	x86PUSHA,      x86POPA,        x86BOUND_GvMa,  x86ARPL_EwGw,    x86Seg_FS,      x86Seg_GS,       x86OpSize,      x86AddrSize,		// 6xh
	x86PUSH_Iv,    x86IMUL_GvEvIv, x86PUSH_Ib,     x86IMUL_GvEvIb,  x86INSB_YbDX,   x86INSWD_YvDX,   x86OUTSB_DXXb,  x86OUTSWD_DXXv,
	x86JO,         x86JNO,         x86JB,          x86JNB,          x86JZ,          x86JNZ,          x86JBE,         x86JNBE,			// 7xh
	x86JS,         x86JNS,         x86JP,          x86JNP,          x86JL,          x86JNL,          x86JLE,         x86JNLE,
	x86Grp1_EbIb,  x86Grp1_EvIv,   x86Undefined,   x86Grp1_EvIb,    x86TEST_EbGb,   x86TEST_EvGv,    x86XCHG_EbGb,   x86XCHG_EvGv,		// 8xh
	x86MOV_EbGb,   x86MOV_EvGv,    x86MOV_GbEb,    x86MOV_GvEv,     x86MOV_EwSw,    x86LEA_GvM,      x86MOV_SwEw,    x86POP_Ev,
	x86NOP,        x86XCHG_eAXeCX, x86XCHG_eAXeDX, x86XCHG_eAXeBX,  x86XCHG_eAXeSP, x86XCHG_eAXeBP,  x86XCHG_eAXeSI, x86XCHG_eAXeDI,	// 9xh
	x86CBW,        x86CWD,         x86CALL_Ap,     x86WAIT,         x86PUSHF_Fv,    x86POPF_Fv,      x86SAHF,        x86LAHF,
	x86MOV_ALOb,   x86MOV_eAXOv,   x86MOV_ObAL,    x86MOV_OveAX,    x86MOVSB_XbYb,  x86MOVSWD_XvYv,  x86CMPSB_XbYb,  x86CMPSWD_XvYv,	// Axh
	x86TEST_ALIb,  x86TEST_eAXIv,  x86STOSB_YbAL,  x86STOSWD_YveAX, x86LODSB_ALXb,  x86LODSWD_eAXXv, x86SCASB_ALYb,  x86SCASWD_eAXYv,
	x86MOV_ALIb,   x86MOV_CLIb,    x86MOV_DLIb,    x86MOV_BLIb,     x86MOV_AHIb,    x86MOV_CHIb,     x86MOV_DHIb,    x86MOV_BHIb,		// Bxh
	x86MOV_eAXIv,  x86MOV_eCXIv,   x86MOV_eDXIv,   x86MOV_eBXIv,    x86MOV_eSPIv,   x86MOV_eBPIv,    x86MOV_eSIIv,   x86MOV_eDIIv,
	x86Grp2_EbIb,  x86Grp2_EvIb,   x86RET_Iw,      x86RET,          x86LES_GvMp,    x86LDS_GvMp,     x86MOV_EbIb,    x86MOV_EvIv,		// Cxh
	x86ENTER_IwIb, x86LEAVE,       x86RETF_Iw,     x86RETF,         x86INT_3,       x86INT_Ib,       x86INTO,        x86IRET,
	x86Grp2_Eb1,   x86Grp2_Ev1,    x86Grp2_EbCL,   x86Grp2_EvCL,    x86AAM,         x86AAD,          x86SALC,        x86XLAT,			// Dxh
	x86FPU,        x86FPU,         x86FPU,         x86FPU,          x86FPU,         x86FPU,          x86FPU,         x86FPU,
	x86LOOPNE_Jb,  x86LOOPE_Jb,    x86LOOP_Jb,     x86JCXZ_Jb,      x86IN_ALIb,     x86IN_eAXIb,     x86OUT_IbAL,    x86OUT_IbeAX,		// Exh
	x86CALL_Jv,    x86JMP_Jv,      x86JMP_Ap,      x86JMP_Jb,       x86IN_ALDX,     x86IN_eAXDX,     x86OUT_DXAL,    x86OUT_DXeAX,
	x86LOCK,       x86Undefined,   x86REPNE,       x86REP,          x86HLT,         x86CMC,          x86Grp3_Eb,     x86Grp3_Ev,		// Fxh
	x86CLC,        x86STC,         x86CLI,         x86STI,          x86CLD,         x86STD,          x86Grp4,        x86Grp5
};


bool  in32BitMode = FALSE;
uchar flip32BitModeCount = 0;

//==================================================================================================================
//==================================================================================================================


int x86Undefined(uchar *in, ulong *out, bool *isBranch, ulong *skip)
{
	return 0;
}


int x86SecondByte(uchar *in, ulong *out, bool *isBranch, ulong *skip)
{
	return 0;
}


int x86OpSize(uchar *in, ulong *out, bool *isBranch, ulong *skip)
{
	// Flip 'in32BitMode' flag.
	in32BitMode = !in32BitMode;

	// Set counter to flip it back after the next opcode...
	flip32BitModeCount = 2;

	return 1;
}


int x86AddrSize(uchar *in, ulong *out, bool *isBranch, ulong *skip)
{
	return 0;
}


int x86Grp1_EbIb(uchar *in, ulong *out, bool *isBranch, ulong *skip)
{
	return 0;
}


int x86Grp1_EvIv(uchar *in, ulong *out, bool *isBranch, ulong *skip)
{
	return 0;
}


int x86Grp1_EvIb(uchar *in, ulong *out, bool *isBranch, ulong *skip)
{
	return 0;
}


int x86Grp2_EbIb(uchar *in, ulong *out, bool *isBranch, ulong *skip)
{
	return 0;
}


int x86Grp2_EvIb(uchar *in, ulong *out, bool *isBranch, ulong *skip)
{
	return 0;
}


int x86Grp2_Eb1(uchar *in, ulong *out, bool *isBranch, ulong *skip)
{
	return 0;
}


int x86Grp2_Ev1(uchar *in, ulong *out, bool *isBranch, ulong *skip)
{
	return 0;
}


int x86Grp2_EbCL(uchar *in, ulong *out, bool *isBranch, ulong *skip)
{
	return 0;
}


int x86Grp2_EvCL(uchar *in, ulong *out, bool *isBranch, ulong *skip)
{
	return 0;
}


int x86Grp3_Eb(uchar *in, ulong *out, bool *isBranch, ulong *skip)
{
	return 0;
}


int x86Grp3_Ev(uchar *in, ulong *out, bool *isBranch, ulong *skip)
{
	return 0;
}


int x86Grp4(uchar *in, ulong *out, bool *isBranch, ulong *skip)
{
	return 0;
}


int x86Grp5(uchar *in, ulong *out, bool *isBranch, ulong *skip)
{
	return 0;
}


//==================================================================================================================
//==================================================================================================================

int x86CompileBlock(uchar *in, ulong *out, int *outLen)
{
	bool isBranch = FALSE;
	uchar *initialIn = in;
	ulong *initialOut = out, skip;

	while (isBranch == FALSE)
	{
		skip = 0;
		in  += x86OpcodeTable[(uchar)*in](in, out, &isBranch, &skip);
		out += skip;

		if (flip32BitModeCount)
		{
			if ((--flip32BitModeCount) == 0)
			{
				// Flip 'in32BitMode' flag back to where it was before...
				in32BitMode = !in32BitMode;
			}
		}
	}

	*outLen = (out - initialOut) << 2;
	return (in - initialIn);
}


#if (PC_VERSION)
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>

void main(int argc, char *argv[])
{
	char *infile, *outfile;
	uchar *inbuf;
	ulong *outbuf;
	FILE *fp;
	int size, xlated, t;
	bool bigEndian;
	ulong *cptr, temp;

/*
	__asm {
		mov ebx, 0x7FFDF00F
		inc ebx
		inc ebx
		mov ebx, 0x7FFDF01F
		inc ebx
		dec ebx
		dec ebx
		inc ebx
		inc ebx
	};
*/

	printf("PS2PC: An Intel x86 emulator for MIPS R5900 (Sony PS2).\n");
	printf("Coded by Mike Tedder / breakpoint - http://ps2pc.parodius.com\n\n");

	if (argc < 3)
	{
		printf("Usage: PS2PC [-e] [-3] <infile> <outfile>\n\n");
		printf("Where <infile> is a file containing a x86 code fragment and <outfile> is a\n");
		printf("file to write the resultant R5900 code to.\n\n");
		printf("Options:\n");
		printf("    -e: write out a big endian version of the code (for NIEW64)\n");
		printf("    -3: assume 32-bit mode (EAX, EBX, etc. support)\n");
		printf("\n");
		return;
	}

	// Init vars.
	infile    = NULL;
	outfile   = NULL;
	bigEndian = FALSE;

	// Parse command line...
	printf("PS2PC: Parsing command line...\n");

	for (t = 1; t < argc; t++)
	{
		if (argv[t][0] == '-')
		{
			switch (toupper(argv[t][1]))
			{
				case 'E':
					printf("PS2PC: Big-endian format enabled.\n");
					bigEndian = TRUE;
					break;

				case '3':
					printf("PS2PC: 32-bit mode enabled.\n");
					in32BitMode = TRUE;
					break;

				default:
					printf("PS2PC: Unknown option '%s'!\n", argv[t]);
					return;
			}
		}
		else
		{
			// Not an option, so assume infile/outfile...
			if (infile == NULL)
				infile = argv[t];
			else
			{
				if (outfile == NULL)
					outfile = argv[t];
				else
				{
					printf("PS2PC: Unknown argument '%s'!\n", argv[t]);
					return;
				}
			}
		}
	}
	
	printf("PS2PC: Loading code from %s...\n", infile);

	fp = fopen(infile, "rb");
	if (fp == NULL)
	{
		printf("PS2PC: Can't open input code!\n");
		return;
	}

	// Seek to eof, get code size.
	fseek(fp, 0, 2);
	size = ftell(fp);
	rewind(fp);

	// Alloc space for code and read it in.
	inbuf = (uchar *)malloc(size);
	if (inbuf == NULL)
	{
		printf("PS2PC: Can't alloc space for input code!\n");
		fclose(fp);
		return;
	}

	fread(inbuf, 1, size, fp);
	fclose(fp);

	// Alloc space for output code.
	outbuf = (ulong *)malloc(size * 32);
	if (outbuf == NULL)
	{
		printf("PS2PC: Can't alloc space for output code!\n");
		free(inbuf);
		return;
	}

	printf("PS2PC: Compiling code...\n");

	xlated = x86CompileBlock(inbuf, outbuf, &size);

	printf("PS2PC: Hit a branch, done compiling.\n");
	printf("PS2PC: Xlated %d bytes of x86 opcodes, resultant R5900 size is %d bytes\n", xlated, size);

	// Flip endian, if requested...
	if (bigEndian)
	{
		printf("PS2PC: Converting to big endian output...\n");

		cptr = (ulong *)outbuf;
		for (t = 0; t < (size >> 2); t++)
		{
			 temp = *cptr;
			*cptr = ((temp & 0xFF000000) >> 24) | ((temp & 0x00FF0000) >> 8) | ((temp & 0x0000FF00) << 8) | ((temp & 0x000000FF) << 24);
			 cptr++;
		}
	}

	printf("PS2PC: Writing output code to %s...\n", outfile);

	fp = fopen(outfile, "wb");
	fwrite(outbuf, 1, size, fp);
	fclose(fp);

	printf("PS2PC: Done!\n");
}
#endif
